{-| Functions to compile agdaLight files to LaTeX documents.
-}

module Compilers.Latex where

import Prelude hiding (print, putStr, putStrLn)
import IO.Output

import Control.Monad.State
import Control.Monad.Writer
import Data.Generics
import Data.List
import Data.Char

import Syntax.Internal
import Syntax.ConcreteToInternal
import Syntax.Lexer
import Syntax.Parser
import Syntax.ParseMonad
import Syntax.Position

import Interface.Options

import Utils.Monads
import Utils.Lists
import Utils.Generics
import Utils.Pairs

-- Mapping the internal categories of names back to the token stream ------

data IName  = IVar Name | IDef Name | ICon Name
	    | IData Name | IField Name
    deriving Show

getName :: IName -> Name
getName (IVar x) = x
getName (IDef x) = x
getName (ICon x) = x
getName (IData x) = x
getName (IField x) = x

-- TODO: do we need this?
setRange r (IVar x)	= IVar x
setRange r (IDef x)	= IDef x
setRange r (ICon x)	= ICon x
setRange r (IData x)	= IData x
setRange r (IField x)	= IField x

instance HasRange IName where
    getRange (IVar x)	= getRange x
    getRange (IDef x)	= getRange x
    getRange (ICon x)	= getRange x
    getRange (IData x)	= getRange x
    getRange (IField x) = getRange x

value x = case p of
	    NoPos   -> (-1,-1)
	    _	    -> (posLine p, posCol p)
    where
	p = rStart (getRange x)

instance Eq IName where
    x == y = (getName x, value x) == (getName y, value y)

instance Ord IName where
    compare x y = compare (value x) (value y)

internalNames :: Data a => a -> [IName]
internalNames =
    everythingBut   (++)
		    isString
		    ([] `mkQ` term `extQ` local
			`extQ` top `extQ` proj
			`extQ` pat `extQ` tel
			`extQ` clause
		    )
    where
	term :: Term -> [IName]
	term t =
	    case t of
		Var r x		-> [IVar x]
		Def r x		-> [IDef x]
		Con r x		-> [ICon x]
		Data r x	-> [IData x]
		Sig r x		-> [IData x]
		Sigma r x _ _	-> [IVar x]
		Struct _ xs	-> map (IField . pi1) xs
		    where pi1 (x,_,_,_) = x
		_		-> []

	local :: LocalDefinition -> [IName]
	local (LocalDef r x (Just _) r' e)	= [IVar x, IVar x]
	local (LocalDef r x _ _ _)		= [IVar x]

	proj (Field r x)	    = [IField $ Name r x 0]
	proj _			    = []

	top :: Definition -> [IName]
	top (NormalDef r x _ _ _ _)	= [IDef x]
	top (Synonym r x _)		= [IDef x]
	top (NewAxiom r x _ _)		= [IDef x]
	top (NewData _ r x _ _ _ cs)	= IData x : map constr cs
	top (NewSig r x _ _ _ ps cs)	= IData x : map IData ps ++ map field cs
	top _				= []

	clause (Clause r x _ _) = [IDef x]

	constr :: Constructor -> IName
	constr (Constructor r x _ _)  = ICon x

	field :: Constructor -> IName
	field (Constructor r x _ _)  = IField x

	pat :: Pattern -> [IName]
	pat (VarP x)	    = [IVar x]
	pat (ConP r x _ _)  = [ICon x]
	pat _		    = []

	tel :: (Name, Term) -> [IName]
	tel (x,t) = [IVar x]

getInternalNames :: Program -> [IName]
getInternalNames p = 
    map head
	$ group
	$ sort
	$ filter (\x -> getName x /= dontCare)
	$ internalNames p

-- Fancy tokens -----------------------------------------------------------

data Space  = Box
	    | Fixed Int
    deriving Show

data FancyToken	= SimpleTok Token
		| IdTok IName
		| OpTok IName
		| SpaceTok Range Space

instance Show FancyToken where
    show t = show (getRange t) ++ " " ++ tokString t

instance HasRange FancyToken where
    getRange (SimpleTok t)  = getRange t
    getRange (IdTok x)	    = getRange x
    getRange (OpTok x)	    = getRange x
    getRange (SpaceTok r _) = r

formatTokens :: [IName] -> [Token] -> [FancyToken]
formatTokens _ [] = []
formatTokens [] ts = map SimpleTok ts
formatTokens ns (t@(TkId (_, "_")) : ts) =
    SimpleTok t : formatTokens ns ts
formatTokens (iname:ns) (t@(TkId (r, x)) : ts)
    | nameString (getName iname) == x
		= IdTok (setRange r iname) : formatTokens ns ts
    | otherwise	= formatTokens ns (t:ts)
--     | otherwise		    = error $ "ns = " ++ show (iname:ns)
-- 				    ++ "\n" ++ replicate 50 '-'
-- 				    ++ "\nts = " ++ show (t : ts)
formatTokens (iname:ns) (t@(TkAtom (r, x)) : ts)
    | nameString (getName iname) == "\"" ++ x ++ "\""
		= IdTok (setRange r iname) : formatTokens ns ts
    | otherwise	= formatTokens ns (t:ts)

--     | otherwise		    = error $ "ns = " ++ show (iname:ns)
-- 				    ++ "\n" ++ replicate 50 '-'
-- 				    ++ "\nts = " ++ show (t : ts)

formatTokens (iname:ns) (t@(TkOp (r, x)) : ts)
    | nameString (getName iname) == x
		= OpTok (setRange r iname) : formatTokens ns ts
    | otherwise	= formatTokens ns (t:ts)
--     | otherwise		    = error $ "ns = " ++ show (iname:ns)
-- 				    ++ "\n" ++ replicate 50 '-'
-- 				    ++ "\nts = " ++ show (t : ts)
formatTokens ns (tp@(TkDot _) : tn@(TkInt (_, n)) : ts)
    | n == 1 || n == 2	= map SimpleTok [tp,tn] ++ formatTokens ns ts
formatTokens ns (t@(TkInt _) : ts) =
	SimpleTok t : formatTokens (dropNat ns) ts
    where
	dropNat (x : xs)
	    | nameString (getName x) == "zero"   = xs
	    | nameString (getName x) == "suc"    = dropNat xs
	    | otherwise		    = error $ "Expected 'suc' or 'zero', found " ++ show x
	dropNat _ = error "impossible: Compilers.Latex.dropNat"

formatTokens ns (t@(TkOpenSquare _) : ts) =
	SimpleTok t : formatTokens (dropList ns) ts
    where
	dropList (x : xs)
	    | nameString (getName x) == "nil"   = xs
	    | nameString (getName x) == ":"	= dropList xs
	    | otherwise		    = error $ "Expected 'nil' or ':', found " ++ show x
	dropList _ = error "impossible: Compilers.Latex.dropList"

formatTokens ns (TkVOpenBrace _ : ts) = formatTokens ns ts
formatTokens ns (TkVCloseBrace _ : ts) = formatTokens ns ts
formatTokens ns (TkVSemi _ : ts) = formatTokens ns ts
formatTokens ns (t : ts)
    | isInfix t	= SimpleTok t : dropInfix ts
    where
	isInfix (TkInfixR _)	= True
	isInfix (TkInfixL _)	= True
	isInfix (TkInfix _)	= True
	isInfix _		= False

	dropInfix (t@(TkInt _) : ts)	= SimpleTok t : dropOps ts
	dropInfix _ = error "impossible: Compilers.Latex.dropInfix"

	dropOps (t@(TkOp _) : ts) =
	    SimpleTok t : dropMoreOps ts
	dropOps (t0@(TkBackQuote _) : t@(TkId _) : t1@(TkBackQuote _) : ts) =
	    map SimpleTok [t0, t, t1] ++ dropMoreOps ts
	dropOps _ = error "impossible: Compilers.Latex.dropOps"

	dropMoreOps (t@(TkComma _) : ts) =
	    SimpleTok t : dropOps ts
	dropMoreOps ts	= formatTokens ns ts

formatTokens ns (t:ts) = SimpleTok t : formatTokens ns ts

-- Texifying --------------------------------------------------------------

texify :: [FilePath] -> IO String
texify files =
    do	let file = last files
	ts <- parseStuff tokensParser file
	ps <- smash <$> mapM (parseStuff programParser) files
	case (ts,ps) of
	    (ParseOk _ ts, ParseOk _ ps)    ->
		do  --putStr $ unlines $ map show ts
		    ips <- case runConvM $ convertPrograms ps of
				Left err    -> fail err
				Right ips   -> return ips
		    --print ips
		    --let inames = getInternalNames $ last ips
		    --putStr $ unlines $ map show $ inames
		    let out0	= phase0 (last ips) ts
			out1	= phase1 out0
			out2	= phase2 out1
			out3	= phase3 out2
			out4	= phase4 out3
			out5	= phase5 out4
			out6	= phase6 out5
-- 			disp n out =
-- 			    do	putStrLn $ "-- Phase " ++ show n ++ " " ++ replicate 50 '-'
-- 				print out
-- 		    disp (-1) ts
-- 		    disp 0 out0
-- 		    disp 1 out1
-- 		    disp 2 out2
-- 		    disp 3 out3
-- 		    disp 4 out4
-- 		    disp 5 out5
		    return out6
	    _	-> error "Parse error (run without -t to get a nice message)"
    where
	convertPrograms [] = return []
	convertPrograms (p:ps) =
	    uncurry (:) <$> programToInternalM p (convertPrograms ps)

	-- smash ParseResults together
	smash = foldr cat (ParseOk undefined [])
	    where
		cat (ParseOk s x) (ParseOk _ xs)    = ParseOk s (x:xs)
		cat (ParseFailed r x y) _	    = ParseFailed r x y
		cat _ (ParseFailed r x y)	    = ParseFailed r x y

	parseStuff p file
	    | isLiterate file	= parseLiterateFile p False file
	    | otherwise		= parseFile p False file

-- New texify algorithm ---------------------------------------------------

data Thing t	= TeX String
		| Agda [t]

instance Show t => Show (Thing t) where
    show (TeX s)    = "TeX\n" ++ unlines (map ("  "++) $ lines s)
    show (Agda ts)  = "Agda\n" ++ unlines (map ("  "++)
				    $ lines $ concatMap show ts)

newThing :: ([a] -> [b]) -> Thing a -> Thing b
newThing f (TeX s)	= TeX s
newThing f (Agda ts)	= Agda $ f ts

newThings = map . newThing

-- Phase 0: Add format directives to tokens

phase0 :: Program -> [Token] -> [FancyToken]
phase0 p = formatTokens (getInternalNames p)

-- Phase 1: Separating Agda from TeX

phase1 :: [FancyToken] -> [Thing FancyToken]
phase1 ts = map toThing1 $ chunks isTeX ts
    where
	isTeX (SimpleTok (TkTeX _)) = True
	isTeX _			    = False

	getTeX (SimpleTok (TkTeX s))	= s
	getTeX _ = error "impossible: Compilers.Latex.getTeX"

	toThing1 ts@(t:_)
	    | isTeX t	= TeX $ concatMap getTeX ts
	    | otherwise	= Agda ts
	toThing1 _ = error "impossible: Compilers.Latex.toThing1"

-- Phase 2: Splitting Agda tokens into lines

type Col    = Int
data Line   = Line Col [FancyToken]

instance Show Line where
    show (Line c ts) = "Line " ++ show c ++ "\n" ++
			unlines (map ("  "++) $ map show ts)

phase2 :: [Thing FancyToken] -> [Thing Line]
phase2 = newThings (split 0)
    where
	split _ []	= []
	split 0 ts	= split (tokLine $ head ts) ts
	split line ts	=
	    case ts0 of
		[]  -> Line 1 [] : split (tokLine $ head ts1) ts1
		_   -> Line (tokCol $ head ts0) ts0 : split (line + 1) ts1
	    where
		(ts0,ts1) = span (\t -> tokLine t == line) ts

	tokLine t = posLine $ rStart $ getRange t
	tokCol	t = posCol  $ rStart $ getRange t

-- Phase 3: Insert explicit spaces between tokens.
--	    (a box space is inserted for indented lines).
--	    After this phase the end position of a token
--	    should equal the start position of the next
--	    token on the same line.

phase3 :: [Thing Line] -> [Thing Line]
phase3 = newThings (map insertSpace)
    where
	insertSpace (Line c ts) = Line c $ box ++ iSpace ts
	    where
		box | c == 1	= []
		    | otherwise	= [SpaceTok r Box]
		start = rStart $ getRange $ head ts
		r   = Range (start { posCol = 1 }) start

	iSpace (t1:t2:ts)
	    | getEnd t1 == getStart t2	= t1 : iSpace (t2:ts)
	    | otherwise			= t1 : SpaceTok r (Fixed 1) : iSpace (t2:ts)
	    where
		r = Range (rEnd $ getRange t1) (rStart $ getRange t2)
	iSpace [t]  = [t, SpaceTok r (Fixed 0)]
	    where
		end = rEnd $ getRange t
		r = Range end end
	iSpace []   = []

	getStart    = posCol . rStart . getRange
	getEnd	    = posCol . rEnd . getRange

-- Phase 4: Turn the lines into a tree shape

-- | The parent of a line @l@ is the closest line above @l@
--   that is indented less than @l@. Empty lines doesn't count.
data LineTree = TreeLine Col [FancyToken] [LineTree]

instance Show LineTree where
    show (TreeLine c ts ls)
	= "Line " ++ show c ++ "\n" ++
	    unlines (map ("  "++) $ map show ts) ++ "\n" ++
	    unlines (map ("  "++) $ lines $ unlines $ map show ls)

phase4 :: [Thing Line] -> [Thing LineTree]
phase4 = newThings buildTree
    where
	buildTree [] = []
	buildTree (Line c [] : ls) = TreeLine c [] [] : buildTree ls
	buildTree (Line c ts : ls) =
		TreeLine c ts (buildTree ls0) : buildTree ls1
	    where
		(ls0, ls1) = span (\l -> isEmpty l || getIndent l > c) ls

	isEmpty (Line _ ts) = null ts

	getIndent (Line c _) = c

-- Phase 5: Computing line indentations and boxes

data BoxedLine	= BLine [[FancyToken]] [FancyToken] [Child]
type BoxIndex	= Int
type Indent	= Int
data Child	= Child BoxIndex Indent BoxedLine

instance Show Child where
    show (Child n i l) = "box" ++ show n ++ "+" ++ show i ++ " -> " ++ show l

instance Show BoxedLine where
    show (BLine boxs ts ls) =
	    "BLine\n" ++
		unlines (map ("  "++) $ concatMap showBox boxs) ++
		"  rest\n" ++
		unlines (map ("    "++) $ map show ts) ++
		unlines (map ("  "++) $ lines $ unlines $ map show ls)
	where
	    showBox ts = "box" : map ("  "++) (map show ts)

phase5 :: [Thing LineTree] -> [Thing BoxedLine]
phase5 = newThings (map computeBoxes)
    where

	computeBoxes (TreeLine _ ts [])	= BLine [] ts []
	computeBoxes (TreeLine c ts ls) =
		BLine boxs ts' $ zipWith mkChild nboxs ls
	    where
		is	= map getIndent ls

		mkChild (n,i) l = Child n i $ computeBoxes l

		-- the box indices for the children
		nboxs	= map findBox is

		findBox c = findBox' c $ zip [0..] ws
		    where
			findBox' c []		    = (-1, c - 1)
			findBox' c ((n, w):_)
			    | w > c		    = (n - 1, c - 1)
			findBox' c [(n,w)]	    = (n, c - w)
			findBox' c ((n, w) : boxs@((_,w') : _))
			    | w' > c	= (n, w - c)
			    | otherwise	= findBox' c boxs

		-- is is decreasing, which means that
		-- cs will be strictly increasing.
		cs  = map head $ group $ reverse is
		tss = splitTokens cs ts
		(wboxs,ts') = getBoxes tss
		(ws, boxs) = unzip wboxs

	-- Compute the boxes and their width from the chopped up token list
	getBoxes :: [[FancyToken]] -> ([(Int, [FancyToken])], [FancyToken])
	getBoxes [ts] = ([], ts)
	getBoxes (box:ts@((t:_):_))   = ((w,box):boxs, ts')
	    where
		w   = posCol $ rStart $ getRange t
		(boxs, ts') = getBoxes ts
	getBoxes tss	= error $ show tss

	getIndent (TreeLine c _ _) = c

	-- Split a list of tokens at given columns
	-- Should only produce the empty list as the last element of the result.
	splitTokens [] ts = [ts]
	splitTokens (c:cs) ts = ts0 `cons` splitTokens cs ts1
	    where
		cons [] xss = xss
		cons xs xss = xs : xss

		(ts0,ts1) = span (endsBefore c) ts
		
				-- The eol token shouldn't be included in a box
		endsBefore c (SpaceTok r (Fixed 0)) = False
		endsBefore c t = posCol (rEnd $ getRange t) <= c

	getNextIndents = getIndents maxBound
	    where
		getIndents prev (Line _ [] : ls) = getIndents prev ls
		getIndents _ (Line 1 _ : _) = []
		getIndents prev (Line c _ : ls)
		    | c < prev	= c : getIndents c ls
		    | otherwise	= [c]
		getIndents _ [] = []

-- Phase 6: Generating TeX

phase6 :: [Thing BoxedLine] -> String
phase6 = unlines . concatMap thingToTeX
    where
	thingToTeX (TeX s)	= [s]
	thingToTeX (Agda ls)	=
	    [ beginAgda ]
	    ++ concatMap blineToTeX ls ++
	    [ endAgda ]
	
	blineToTeX (BLine [] [] []) =
	    [ mathMode "{}" ++ par ]
	blineToTeX (BLine boxs ts ls) =
		[ "{" ]
		++ map ("  "++) (zipWith setbox [0..] boxs'
		    ++ [ mathMode (copy ++ tokensToTeX ts') ++ par ]
		    ++ concatMap childToTeX ls
		    )
		++ [ "}" ]
	    where
		copy = case length boxs of
			    0	-> ""
			    n	-> copyBox (n - 1)

		setbox 0 box = setBox 0 $ mathMode $ tokensToTeX box
		setbox n box = setBox n	$ mathMode
					$ copyBox (n - 1) ++
					  tokensToTeX box

		(boxs',ts') = initLast $ mathSpace (boxs ++ [ts])

{-
		box' = mathSpace box
		ts' | null box		    = mathSpace' ts
		    | isMathRel (last box)  = mathSpace ts
		    | otherwise		    = mathSpace' ts
-}
	childToTeX (Child n i l) = saveWidth n i : blineToTeX l

	tokensToTeX ts = concatMap tokString ts

	isSpace (SpaceTok _ _)   = True
	isSpace _		    = False

	initLast [x] = ([], x)
	initLast (x:xs) = (x : xs', y)
	    where (xs',y) = initLast xs
	initLast _ = error "impossible: Compilers.Latex.initLast"

	headCons x (xs:xss) = (x:xs) : xss
	headCons _ _ = error "impossible: Compilers.Latex.headCons"

	mathSpace tss = skipIndent tss
	    where
		skipIndent ((t:ts):tss)
		    | isSpace t	    = t `headCons` mathSpace' (ts:tss)
		    | otherwise	    = mathSpace' ((t:ts):tss)
		skipIndent _ = error "impossible: Compilers.Latex.skipIndent"

	mathSpace' ([]:tss) = [] : mathSpace' tss
	mathSpace' ([t]:(t':ts):tss)
	    | isSpace t && isMathRel t'	= [] : mathSpace' ((t':ts):tss)
	    | isMathRel t && isSpace t'	= [t] : mathSpace' (ts:tss)
	    | otherwise			= [t] : mathSpace' ((t':ts):tss)
	mathSpace' ((t:t':ts):tss)
	    | isSpace t && isMathRel t'	= mathSpace' ((t':ts):tss)
	    | isMathRel t && isSpace t'	= mathSpace' ((t:ts):tss)
	    | otherwise			= t `headCons` mathSpace' ((t':ts):tss)
	mathSpace' [[t]]    = [[t]]
	mathSpace' []	    = []
	mathSpace' tss	    = error $ show tss

isMathRel t =
    case t of
	SimpleTok (TkArrow _)	-> True
	SimpleTok (TkColon _)	-> True
	SimpleTok (TkEqual _)	-> True
	SimpleTok (TkOp _)	-> True
	SimpleTok (TkDollar _)	-> True
	_			-> False

tokString (SpaceTok _ Box) = boxSpace
tokString (SpaceTok _ (Fixed n)) = concat $ replicate n singleSpace
tokString (IdTok n) =
    case n of
	IDef x	    -> definition $ escapeName x
	IVar x	    -> variable $ escapeName x
	IData x	    -> datatype $ escapeName x
	IField x    -> field $ escapeName x
	ICon x	    -> constructor $ escapeName x
tokString (OpTok n) =
    case n of
	IDef x	    -> definition $ escapeOp x
	IVar x	    -> variable $ escapeOp x
	IData x	    -> datatype $ escapeOp x
	IField x    -> field $ escapeOp x
	ICon x	    -> constructor $ escapeOp x
tokString (SimpleTok t) =
    case t of
	TkLambda _	    -> "\\lambda{}"
	TkArrow _	    -> "{}\\to{}"
	TkFrom _	    -> "{}\\leftarrow{}"
	TkId (_,x)	    -> escapeName $ dummyName x
	TkAtom (_,x)	    -> escapeName $ dummyName x
	TkLParen _	    -> "("
	TkRParen _	    -> ")"
	TkSemi _	    -> ";"
	TkVSemi _	    -> ""
	TkComma _	    -> ","
	TkUniverse (_,n)    -> newsort $ "\\#" ++ show n
	TkInfix _	    -> keyword "infix"
	TkInfixL _	    -> keyword "infixl"
	TkInfixR _	    -> keyword "infixr"
	TkBackQuote _	    -> "`"
	TkRewrite _	    -> keyword "rewrite"
	TkDo _		    -> keyword "do"
	TkWhere _	    -> keyword "where"
	TkWith _	    -> keyword "with"
	TkMutual _	    -> keyword "mutual"
	TkAbstract _	    -> keyword "abstract"
	TkClass _	    -> keyword "class"
	TkInstance _	    -> keyword "instance"
	TkData _	    -> keyword "data"
	TkSized _	    -> keyword "sized"
	TkExtends _	    -> keyword "extends"
	TkInt (_,n)	    -> show n
	TkDblArrow _	    -> "{}\\Rightarrow{}"
	TkOpenBrace _	    -> "\\{"
	TkCloseBrace _	    -> "\\}"
	TkOpenSquare _	    -> "["
	TkCloseSquare _	    -> "]"
	TkEqual _	    -> "{}={}"
	TkDot _		    -> "."
	TkSig _		    -> keyword "sig"
	TkStruct _	    -> keyword "struct"
	TkLet _		    -> keyword "let"
	TkIn _		    -> keyword "in"
	TkOf _		    -> keyword "of"
	TkPlugin (_,x)	    -> keyword $ x ++ "{-}plugin"
	TkUnderscore _	    -> "\\_"
	TkColon _	    -> ":"
	TkDollar _	    -> "{}\\mathrel{\\$}{}"
	TkOp (_, x)	    -> escapeOp $ dummyName x
	TkVOpenBrace _	    -> ""
	TkVCloseBrace _	    -> ""
	TkSet _		    -> newsort "Set"
	TkISet _	    -> newsort "ISet"
	TkProp _	    -> newsort "Prop"
	TkType _	    -> newsort "Type"
	TkStar _	    -> newsort "Star"
	TkTeX _		    -> error "impossible: Compilers.Latex.tokString TkTeX"
	TkEOF		    -> error "impossible: Compilers.Latex.tokString TkEOF"

keyword x	= "\\AgdaKeyword{" ++ x ++ "}"

variable x	= "\\AgdaVariable{" ++ x ++ "}"
definition x	= "\\AgdaDefinition{" ++ x ++ "}"
constructor x	= "\\AgdaConstructor{" ++ x ++ "}"
datatype x	= "\\AgdaDatatype{" ++ x ++ "}"
newsort x	= "\\AgdaSort{" ++ x ++ "}"
field x		= "\\AgdaField{" ++ x ++ "}"

specialOps =
    [ "<="	|-> "{}\\leq{}"
    , "/="	|-> "{}\\neq{}"
    , ">="	|-> "{}\\geq{}"
    , "==>"	|-> "{}\\Longrightarrow{}"
    , "<=="	|-> "{}\\Longleftarrow{}"
    , "/\\"	|-> "{}\\wedge{}"
    , "\\/"	|-> "{}\\vee{}"
    , "<=>"	|-> "{}\\Leftrightarrow{}"
    , "<==>"	|-> "{}\\Longleftrightarrow{}"
    , "-->"	|-> "{}\\longrightarrow{}"
    , "<--"	|-> "{}\\longleftarrow{}"
    , "<-->"	|-> "{}\\longleftrightarrow{}"
    , "<->"	|-> "{}\\leftrightarrow{}"
    , "<-"	|-> "{}\\leftarrow{}"
    , "|-"	|-> "{}\\vdash{}"
    , "-|"	|-> "{}\\dashv{}"
    , "|="	|-> "{}\\models{}"
    , "<>"	|-> "{}\\Diamond{}"
    , "<<"	|-> "{}\\ll{}"
    , ">>"	|-> "{}\\gg{}"
    , "~>"	|-> "{}\\leadsto{}"
    , "|->"	|-> "{}\\mapsto{}"
    , "|-->"	|-> "{}\\longmapsto{}"
    , "||"	|-> "{}\\parallel{}"
    , "|><|"	|-> "{}\\Join{}"
    , "@"	|-> "{}\\circ{}"
    , "<:"	|-> "{}\\subseteq{}"
    , ":>"	|-> "{}\\supseteq{}"
    , "~="	|-> "{}\\cong{}"
    , "~~"	|-> "{}\\approx{}"
    , "~"	|-> "{}\\sim{}"
    , "~-"	|-> "{}\\simeq{}"
    , "=."	|-> "{}\\doteq{}"
    ]
    where (|->) = (,)

escapeOp DontCare     = error "impossible: Compilers.Latex.escapeOp DontCare"
escapeOp (Name _ x _) = maybe (mkOp x) id (lookup x specialOps)
    where
	mkOp x = "{}\\mathrel{" ++
		    concat (narrow $ intersperse "\\mkern-4mu" $ map esc x)
		    ++ "}{}"

	esc '&'	    = "\\&"
	esc '%'	    = "\\%"
	esc '\\'    = "\\backslash{}"
	esc '$'	    = "\\$"
	esc '^'	    = "\\hat{}"
	esc c	    = [c]

	narrow (_:"|":xs)   = "|" : narrow xs
	narrow (_:"!":xs)   = "!" : narrow xs
	narrow (_:"'":xs)   = "'" : narrow xs
	narrow (x:xs)	    = x : narrow xs
	narrow []	    = []

specialNames = []

specialChars =
    [ '_'	|-> "\\_"
    , '\x393'	|-> "\\Gamma{}"
    , '\x394'	|-> "\\Delta{}"
    , '\x398'	|-> "\\Theta{}"
    , '\x39e'	|-> "\\Xsi{}"
    , '\x3a0'	|-> "\\Pi{}"
    , '\x3a5'	|-> "\\Upsilon{}"
    , '\x3a6'	|-> "\\Phi{}"
    , '\x3a8'	|-> "\\Psi{}"
    , '\x3a9'	|-> "\\Omega{}"
    , '\x3b1'	|-> "\\alpha{}"
    , '\x3b2'	|-> "\\beta{}"
    , '\x3b3'	|-> "\\gamma{}"
    , '\x3b4'	|-> "\\delta{}"
    , '\x3b5'	|-> "\\epsilon{}"
    , '\x3b6'	|-> "\\zeta{}"
    , '\x3b7'	|-> "\\eta{}"
    , '\x3b8'	|-> "\\theta{}"
    , '\x3b9'	|-> "\\iota{}"
    , '\x3ba'	|-> "\\kappa{}"
    , '\x3bb'	|-> "\\lambda{}"
    , '\x3bc'	|-> "\\mu{}"
    , '\x3bd'	|-> "\\nu{}"
    , '\x3be'	|-> "\\xsi{}"
    , '\x3bf'	|-> "o"
    , '\x3c0'	|-> "\\pi{}"
    , '\x3c1'	|-> "\\rho{}"
    , '\x3c2'	|-> "\\varsigma{}"
    , '\x3c3'	|-> "\\sigma{}"
    , '\x3c4'	|-> "\\tau{}"
    , '\x3c5'	|-> "\\upsilon{}"
    , '\x3c6'	|-> "\\phi{}"
    , '\x3c7'	|-> "\\chi{}"
    , '\x3c8'	|-> "\\psi{}"
    , '\x3c9'	|-> "\\omega{}"
    ]
    where
	(|->) = (,)

escapeName (Name _ x _) = maybe (atomic $ concatMap esc stem ++ suf) id (lookup x specialNames)
    where
	x' | isQuoted x = init $ tail x
	   | otherwise	= x
	(ix,stem) = reverse -*- reverse $ span isDigit $ reverse x'
	suf = case ix of
		""  -> ""
		_   -> "_{" ++ ix ++ "}"
	esc c = maybe [c] id $ lookup c specialChars
	isQuoted ('"':_) = True
	isQuoted _	 = False
	atomic y
	    | isQuoted x = "\\mbox{``{\\tt " ++ y ++ "}''}"
	    | otherwise  = y
escapeName DontCare = "\\_"

setBox n s  = "\\setbox" ++ show n ++ "=\\hbox{" ++ s ++ "}"
copyBox	n   = "\\copy" ++ show n
saveWidth (-1) i    = "\\edef\\Indent{" ++
			longSpace i ++ "}"
saveWidth n i	    = "\\edef\\Indent{\\hskip\\wd" ++ show n ++
			longSpace i ++ "}"
boxSpace    = "\\Indent{}"
mathMode "" = "${}$"
mathMode s  = "$" ++ s ++ "$"
par	    = "\\par"

beginAgda = "\\begin{agda}"
endAgda	    = "\\end{agda}"
longSpace 0 = ""
longSpace n = "\\hskip" ++ show (2 * n) ++ "mm"
newline = "{}\\\\\n"
singleSpace = "\\,"

-- HasRange Token

instance HasRange Token where
    getRange t = case t of
	TkLambda r	    -> r
	TkArrow r	    -> r
	TkId (r, _)	    -> r
	TkLParen r	    -> r
	TkRParen r	    -> r
	TkSemi r	    -> r
	TkVSemi r	    -> r
	TkComma r	    -> r
	TkUniverse (r,_)    -> r
	TkStar (r,_)	    -> r
	TkSet r		    -> r
	TkISet r	    -> r
	TkProp r	    -> r
	TkType r	    -> r
	TkInfix r	    -> r
	TkInfixL r	    -> r
	TkInfixR r	    -> r
	TkBackQuote r	    -> r
	TkRewrite r	    -> r
	TkExtends r	    -> r
	TkWhere r	    -> r
	TkWith r	    -> r
	TkMutual r	    -> r
	TkAbstract r	    -> r
	TkClass r	    -> r
	TkInstance r	    -> r
	TkData r	    -> r
	TkDo r		    -> r
	TkFrom r	    -> r
	TkSized r	    -> r
	TkSig r		    -> r
	TkStruct r	    -> r
	TkOf r		    -> r
	TkInt (r, _)	    -> r
	TkDblArrow r	    -> r
	TkOpenBrace r	    -> r
	TkCloseBrace r	    -> r
	TkOpenSquare r	    -> r
	TkCloseSquare r	    -> r
	TkVOpenBrace r	    -> r
	TkVCloseBrace r	    -> r
	TkEqual r	    -> r
	TkColon r	    -> r
	TkDot r		    -> r
	TkDollar r	    -> r
	TkLet r		    -> r
	TkIn r		    -> r
	TkPlugin (r,_)	    -> r
	TkUnderscore r	    -> r
	TkOp (r, _)	    -> r
	TkAtom (r, _)	    -> r
	TkEOF		    -> noRange
	TkTeX _		    -> noRange

